/*
 *
 *
 *    MotorAccel.h
 *
 *    Stepper motor interface, based on AccelStepper library.
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef AUTOCAP_MOTORACCEL_H
#define AUTOCAP_MOTORACCEL_H

// Arduino environment stuff
#include <Arduino.h>

// required library header - you must download this library and install it
//    into your Arduino development environment
#include <AccelStepper.h>

// required local headers
#include "Elapsed.h"
#include "Interfaces.h"
#include "EnablePin.h"

///////////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: Set ACCEL_ENABLE_SUPPORT to 'true' to add support for ENABLE pin
//            management.  This allows the controller to use a PWM output line
//            to control the average current provided to simple H-bridge motor
//            controllers.  It can alternatively be used with a normal digital
//            line to toggle the ENABLE line of a dedicated stepper controller,
//            to put the controller into low-power mode when idle.
//
//            Enabling this feature does require a few more bytes of RAM for
//            each motor port, and it slightly increases the size of the
//            program in FLASH.
//
#define ACCEL_ENABLE_SUPPORT true

///////////////////////////////////////////////////////////////////////////////
//
//  OPTIONAL: For some reason the AccelStepper library doesn't set the coil
//            outputs until a step command is requested.  Setting this to true
//            will cause the Setup() function to 'tickle' the library by
//            stepping forward then back by one step.
//
#define ACCEL_MOTOR_TICKLE true

///////////////////////////////////////////////////////////////////////////////
//
//  class MotorAccel - Stepper Motor Interface
//
//  This is a wrapper around the AccelStepper class, to give it an interface
//  that is consistent with the other motor types.  It also tracks certain
//  runtime state items that are needed to make the AccelStepper object work
//  properly.  Unsupported interface details are likewise hidden.
//
class MotorAccel : public IMotor {
	public:
		// The maximum speed, specified in steps per second.
		float MaxSpeed;

		// the acceleration rate, specified in steps per second^2. To disable, set to (0).
		float Acceleration;

	private:
		AccelStepper Motor;	// the motor object

		enum States {
			Stopped,	// not doing anything
			Moving,		// continuous movement
			Pulsing,	// time-based pulse generation
			Stepping,	// fixed-step movement
			Stopping,	// wait time after last step
		};

		unsigned long PulseStart;	// when the current pulse started
		unsigned long PulseLength;	// when the current pulse should end
		States State;				// current action state
		const uint8_t PinLED;		// the LED pin assignment

		#if ACCEL_ENABLE_SUPPORT == true
		const uint8_t DisableDelay;	// the number of msec to wait before powering down on stop
		EnablePin * const Enables;	// ENABLE pin definitions
		const uint8_t EnableLen;	// number of ENABLE pins supported
		#endif

	private:
		// common ctor code
		void Init() {
			// start with a clean state
			PulseStart = PulseLength = 0UL;
			State = Stopped;
			MaxSpeed = 200.0; // steps/sec
			Acceleration = 0;

			#if ACCEL_ENABLE_SUPPORT == true
			// make sure ENABLE values are sane for the selected options
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				if (ep->DisableOnStop && !ep->IsPWM) {
					if (ep->EnablePWM == 0xFF)
						ep->EnablePWM = HIGH;
					if (ep->DisablePWM == 0x00)
						ep->DisablePWM = LOW;
				}
			}
			#endif
		}

		#if ACCEL_ENABLE_SUPPORT == true
		// clear the optional Enable/PWM lines
		void Disable() {
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				if (ep->DisableOnStop) {
					if (ep->IsPWM) {
						analogWrite(ep->Pin, ep->DisablePWM);
					} else {
						digitalWrite(ep->Pin, ep->DisablePWM);
					}
				}
			}
		}

		// assert the optional Enable/PWM lines
		void Enable() {
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				if (ep->IsPWM) {
					analogWrite(ep->Pin, ep->EnablePWM);
				} else {
					digitalWrite(ep->Pin, ep->EnablePWM);
				}
			}
		}
		#endif

		// EndStop() - complete a stopping action
		void EndStop() {
			#if ACCEL_ENABLE_SUPPORT == true
			Disable();
			#endif

			// clear any running flags
			State = Stopped;
		}

	public:
		//
		//  constructor - arguments match constructor from AccelStepper class
		//
		MotorAccel(void (*forward)(), void (*backward)(), uint8_t pinLed = 0xFF)
			: Motor(forward, backward),
			  #if ACCEL_ENABLE_SUPPORT == true
			  Enables(0), EnableLen(0), DisableDelay(0x00),
			  #endif
			  PinLED(pinLed)
		{
			Init(); // call common ctor code
		}

		#if ACCEL_ENABLE_SUPPORT == false
		//
		//  constructor - arguments match constructor from AccelStepper class
		//
		MotorAccel( uint8_t interface = AccelStepper::FULL4WIRE,
					uint8_t pin1 = 2,
					uint8_t pin2 = 3,
					uint8_t pin3 = 4,
					uint8_t pin4 = 5,
					uint8_t ledPin = 0xFF)
			: Motor(interface, pin1, pin2, pin3, pin4, true), PinLED(ledPin) {
			Init(); // call common ctor code
		}
		#else
		//
		//  constructor - extended versions to support ENABLE pin
		//
		//  AccelStepper Arguments (see docs):
		//     interface     - the AccelStepper mode
		//     pin1          - the first motor control output pin
		//     pin2          - the second motor control output pin
		//     pin3          - the third motor control output pin
		//     pin4          - the fourth motor control output pin
		//
		//  Optional Arguments:
		//     ledPin        - GPIO output for port-select LED
		//     enable        - array of EnablePin definitions
		//     enableCt      - number of items in 'enable' array
		//     disableDelay  - when the motor is coming to a stop, this is the number
		//                     of milliseconds to wait between taking the final step
		//                     and changing ENABLE output to pwmValueOff.
		//
		//  NOTE: When using PWM on the ENABLE output to control voltage/current,
		//        make sure to use a PWM pin that has an appropriately high frequency
		//        output, or modify the appropriate timer to change the PWM output
		//        frequency.  Using a PWM frequency that is too low can cause erratic
		//        motor behavior.
		//
		MotorAccel( uint8_t interface,				// AccelStepper mode (e.g., AccelStepper::FULL4WIRE
					uint8_t pin1,					// first motor control pin (GPIO)
					uint8_t pin2,					// second motor control pin (GPIO)
					uint8_t pin3,					// third motor control pin (GPIO)
					uint8_t pin4,					// fourth motor control pin (GPIO)
					uint8_t ledPin,					// port-select LED pin (GPIO)
					float maxSpeed,					// max steps per second
					EnablePin * const enable = 0,	// array of Enable pin definitions
					const uint8_t enableCt = 0,		// number of Enable pins in the array
					uint8_t disableDelay = 0)		// settling delay between last motor
													//   STEP and PWM STOPPED value (msec)
			: Motor(interface, pin1, pin2, pin3, pin4, true),
			  MaxSpeed(maxSpeed),
			  PinLED(ledPin),
			  Enables(enable),
			  EnableLen(enableCt),
			  DisableDelay(disableDelay) {
			// call common ctor code
			Init();
		}

		MotorAccel( uint8_t interface,				// AccelStepper mode (e.g., AccelStepper::FULL4WIRE
					uint8_t pin1,					// first motor control pin (GPIO)
					uint8_t pin2,					// second motor control pin (GPIO)
					uint8_t pin3,					// third motor control pin (GPIO)
					uint8_t pin4,					// fourth motor control pin (GPIO)
					float maxSpeed,					// max steps per second
					EnablePin * const enable = 0,	// array of Enable pin definitions
					const uint8_t enableCt = 0,		// number of Enable pins in the array
					uint8_t disableDelay = 0)		// settling delay between last motor
													//   STEP and PWM STOPPED value (msec)
			: Motor(interface, pin1, pin2, pin3, pin4, true),
			  MaxSpeed(maxSpeed),
			  PinLED(0xFF),
			  Enables(enable),
			  EnableLen(enableCt),
			  DisableDelay(disableDelay) {
			// call common ctor code
			Init();
		}
		#endif

		//
		//  ~MotorAccel - empty virtual dtor placeholder
		//
		virtual ~MotorAccel() { /* nop */ }

		//
		//  Setup() - initialize the hardware; this should be called once from setup()
		//
		virtual void Setup() {
			// initialize the motor parameters
			Motor.setMaxSpeed(MaxSpeed);
			Motor.setAcceleration(Acceleration);

			// FIXME: for some reason, we have to 'tickle' the library to
			//        get the outputs to enable
			#if ACCEL_MOTOR_TICKLE == true
			Motor.move(1);
			Motor.setSpeed(MaxSpeed);
			while (!Motor.distanceToGo() == 0L)
				Motor.runSpeedToPosition();

			Motor.move(-1);
			Motor.setSpeed(MaxSpeed);
			while (!Motor.distanceToGo() == 0L)
				Motor.runSpeedToPosition();
			#endif

			#if ACCEL_ENABLE_SUPPORT == true
			// initialize Enable pin
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				pinMode(ep->Pin, OUTPUT);
				if (ep->DisableOnStop) {
					if (ep->IsPWM) {
						analogWrite(ep->Pin, ep->DisablePWM);
					} else {
						digitalWrite(ep->Pin, ep->DisablePWM);
					}
				} else {
					if (ep->IsPWM) {
						analogWrite(ep->Pin, ep->EnablePWM);
					} else {
						digitalWrite(ep->Pin, ep->EnablePWM);
					}
				}
			}
			#endif

			// set motor state to 'stopped'
			Stop();
		}
		
		//
		//  Pulse(...) - pulse the motor with a specific direction, speed, and duration
		//
		virtual bool Pulse(bool direction, uint8_t speed, uint16_t howLong) {
			// record the pulse details
			PulseStart = millis();		// get the start stamp
			PulseLength = howLong;		// compute pulse end time

			// use 'Move' to get the motor underway
			Move(direction, speed);
			State = Pulsing;	// but change the state to Pulsing, instead

			// success
			return true;
		}
		
		//
		//  Move(...) - move the motor continuously with a specific direction and speed
		//
		virtual bool Move(bool direction, uint8_t speed) {
			// Move(speed=0) is a synonym for Stop()
			if (speed == 0) {
				Stop();
				return true;
			}

			// compute the speed
			float howFast = MaxSpeed * (((float)speed) / 255.0);
			if (!direction) {
				howFast = -howFast;
			}

			// start the motor moving
			#if ACCEL_ENABLE_SUPPORT == true
			Enable();
			#endif
			Motor.setSpeed(howFast);
			State = Moving;

			// success
			return true;
		}

		//
		//  Step(...) - move step-equipped motor specific number of steps at given rate
		//
		virtual bool Step(bool direction, uint8_t speed, uint16_t steps) {
			// Step(speed=0) is a synonym for Stop()
			if (speed == 0) {
				Stop();
				return true;
			}

			// compute number of steps
			long howFar = steps;
			if (!direction) 
				howFar = -howFar;

			// compute the speed
			float howFast = MaxSpeed * (((float)speed) / 255.0);
			if (!direction) {
				howFast = -howFast;
			}

			// update motor state
			#if ACCEL_ENABLE_SUPPORT == true
			Enable();
			#endif
			Motor.move(howFar);
			Motor.setSpeed(howFast);
			State = Stepping;

			// success
			return true;
		}

		//
		//  Run() - update state, timers, actions, etc.; this should be called once per loop() iteration
		//
		virtual void Run() {
			unsigned long now = millis();

			switch (State) {
				case Stopped:
					// service motor object
					Motor.run();
					break;

				case Moving:
					// service motor object
					Motor.runSpeed();
					break;

				// pulse in progress
				case Pulsing:
					// service motor object
					Motor.runSpeed();

					if (Elapsed(now, PulseStart) >= PulseLength) {
						// command the motor to STOP
						Stop();
					}
					break;

				case Stepping:
					// service motor object
					Motor.runSpeedToPosition();

					if (Motor.distanceToGo() == 0L) {
						// command the motor to STOP
						Stop();
					}

					//
					//  FIXME: we should be able to just call run() when stepping, but for
					//         some reason, this does not work.  The runSpeedToPosition()
					//         works, but doesn't implement acceleration ramps, so if we
					//         want to implement that later, we need to figure out why the
					//         run() function doesn't work here.
					//

					break;

				case Stopping:
					// service motor object
					Motor.run();
					// if enough delay has passed, go ahead and disable the motor
					if (Elapsed(now, PulseStart) >= PulseLength) {
						EndStop();
					}
					break;
			}
		}
		
		//
		//  Stop() - stop the motor, regardless of what it is doing
		//
		virtual bool Stop() {
			// stop all motor activity
			Motor.stop();
			Motor.setSpeed(0);

#if ACCEL_ENABLE_SUPPORT == true
			// move to Stopping state
			State = Stopping;
			// use the pulse timers to record the delay
			PulseStart = millis();
			PulseLength = DisableDelay;
#else
			// disable the motor drive power
			EndStop();
#endif

			// success
			return true;
		}
		
		//
		//  ReadCurrent() - read the motor current and return it
		//
		virtual float ReadCurrent() {
			// not supported
			return 0.0;
		}
		
		//
		//  LED(state) - update the 'port selected' LED state
		//
		virtual void LED(bool state) {
			digitalWrite(PinLED, state);
		}
		
		//
		//  SetBrake(state) - set state of braking feature
		//
		virtual bool SetBrake(bool state) {
			#if ACCEL_ENABLE_SUPPORT == true
			// save the flag; applies only to this port
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				ep->DisableOnStop = !state;
			}

			// FIXME: refactor the EnablePin interface so that the leading Enable() below isn't needed
			if (State == Stopped) {
				Enable();
				Disable();
			}

			// success
			return true;
			#else
			// not supported
			return false;
			#endif
		}

		//
		//  GetBrake() - get the brake setting
		//
		virtual bool GetBrake() {
			#if ACCEL_ENABLE_SUPPORT == true
			// save the flag; applies only to this port
			bool result = false;
			for (uint8_t i = 0; i != EnableLen; ++i) {
				EnablePin * const ep = Enables + i;
				result |= !ep->DisableOnStop;
			}
			#else
			return false;
			#endif
		}

		//
		//  Read the current stepper position; return true if supported
		//
		virtual bool GetPosition(long *outValue) {
			*outValue = Motor.currentPosition();
			return true;
		}

		//
		//  Seek() - go to a specific position
		//
		virtual bool Seek(long target, uint8_t speed) {
			// compute the speed
			float howFast = MaxSpeed * (((float)speed) / 255.0);

			// update motor state
			#if ACCEL_ENABLE_SUPPORT == true
			Enable();
			#endif
			Motor.moveTo(target);
			Motor.setSpeed(fabs(howFast));
			State = Stepping;

			// success
			return true;
		}
		
		//
		//  Reset the current stepper position to zero; return true if supported
		//
		virtual bool ZeroPosition(void) {
			// resetting the counter mid-command breaks the library state; stop first, then reset
			this->Stop();

			// reset the position
			Motor.setCurrentPosition(0);
			return true;
		}
		
		//
		//  Get the named EnablePin object
		//
		virtual EnablePin * GetEnable(uint8_t idx) {
			if (idx >= EnableLen) return 0;
			return Enables + idx;
		}

		//
		//  Returns true iff the motor is moving
		//
		virtual bool IsRunning() const {
			return (State != Stopped) && (State != Stopping);
		}
};

#endif // AUTOCAP_MOTORACCEL_H
