/*
 *
 *
 *    MotorDC.h
 *
 *    Simple DC motor interface.
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

//  ==================================================================
//  NOTES FOR PIN ASSIGNMENTS
//  ==================================================================
//
//  NOTE 0: Any feature can be disabled by setting the corresponding pin number
//          to negative one (-1).
//
//  NOTE 1: The standard Arduino Motor Shield R3 uses these values:
//            http://arduino.cc/en/Main/ArduinoMotorShieldR3
//        NUM_PORTS = 2
//           PORT 1:
//              PWM   = 3
//              BRAKE = 9
//              DIR   = 12
//              CS    = 0 (A0)
//           PORT 2:
//              PWM   = 11
//              BRAKE = 8
//              DIR   = 13
//              CS    = 1 (A1)
//
//          When using the standard shield, the UP/DOWN output pins are not
//          needed, and you can leave them set to zero.  If you want to use
//          those pins to drive some additional LEDs or other indicators,
//          you are certainly free to do so.
//
//  NOTE 2: The Ardumoto Shield from SparkFun also uses the same PWM and
//          Direction pin assignments for both ports, but does not provide
//          the sense or brake functions at all.  So enabling these functions
//          while using this shield model will not produce useful braking,
//          nor meaningful current readings without modifications to the shield.
//
//  NOTE 3: You can also use a relay board in an H-bridge using the UP/DOWN
//          pin assignments, in lieu of a motor shield.  You can always mix
//          and match features, e.g., use UP/DOWN for a relay, but drive
//          a MOSFET with a PWM output, to generate PWM to drive your motor
//          through your relay contacts.
//
//  NOTE 4: Another possible configuration is to use the direction pin to
//          drive a DPDT relay to select direction, and the UP/DOWN pins
//          to drive a second relay or a transistor to turn the power on
//          or off.
//
//  NOTE 5: The UP & DOWN functions can be assigned to the same pin as long
//          as you use the 'direction' pin to actually control direction.
//

#ifndef AUTOCAP_MOTORDC_H
#define AUTOCAP_MOTORDC_H

// required local headers
#include "Elapsed.h"
#include "MotorPins.h"
#include "Interfaces.h"

///////////////////////////////////////////////////////////////////////////////
//
//  class MotorDC - Simple DC Motor Interface
//
class MotorDC : public IMotor {
	private:
		static const int CurrentZero = 10;		// the A/D value for 0.0 mA reading
		static const int BrakePWM = 0xFF;		// PWM value to use when brake enabled(*2)
		static const int BrakePolarity = HIGH;	// polarity to assert when brake is engaged(*1)
		//
		//  Notes(*):
		//     (1) This note only applies to PWM/DIRECTION/BRAKE configurations.
		//         See also the L298P datasheet, and the Arduino Motor Shield R3
		//         schematic.  Asserting the BRAKE line to the shield will cause
		//         both INPUTs of that channel to be of like polarity, which, when
		//         combined with an ENABLE value of HIGH (PWM = 255), will cause
		//         the controller to actively brake the motor.  The value of the
		//         two inputs in this case will be either both high, or both low,
		//         depending on the state of the DIRECTION pin.  Setting the
		//         BrakePolarity = HIGH will cause the braked condition to be a
		//         POSITIVE output on both motor power leads, whereas setting a
		//         value of BrakePolarity = LOW will cause the braked condition
		//         to be a COMMON (ZERO) output on both motor power leads.  Both
		//         configurations (both high, or both low) are braking conditions,
		//         but the both-high condition also enables any shield LEDs that
		//         are installed for the braked port, providing positive feedback
		//         that the port is indeed in a braked condition.  If you prefer a
		//         brake condtion of both-low, simply change BrakePolarity = LOW.
		//     (2) BrakePWM is the PWM value written to the ENABLE input of the
		//         L298P chip for a port to brake it, combined with a BRAKE = HIGH
		//         assertion for that port, and the polarity selection from (1).
		//         The BrakePWM should always be 255 to force a constant HIGH
		//         output on the PWM lead of the Arduino, which will hold the
		//         ENABLE input HIGH on the L298P.  Smaller values will provide
		//         smaller amounts of braking.  A value of zero will disable
		//         braking altogether, regardless of the state of other pins.

	private:
		const MotorPins motorPins;	// pin definitions
		unsigned long PulseStart;	// when the current pulse started
		unsigned long PulseLength;	// when the current pulse should end
		float Current;				// the current (I) reading from the motor
		boolean Pulsing;			// indicates that a pulse is in progress
		boolean Running;			// indicates that a move is in progress
		boolean BrakingEnabled;		// indicates that braking feature is enabled

	private:
		//
		//  setCurrent(...) - Takes a raw A/D reading, converting and storing it
		//                    as a true value, measured in amps.  This only works
		//                    for the current sense built into the Arduino Motor
		//                    Shield R3, or an equivalent circuit.
		//
		void SetCurrent(unsigned raw) {
			Current = (float)(raw - CurrentZero) * 0.00295928; // A / unit
		}
		
		//
		//  CommandMotor(...) - Sets or changes motor state; if the motor port supports
		//						braking, and the command is STOP (effort = 0), then the
		//						brake will be enabled, as well.
		//
		void CommandMotor(bool polarity, uint8_t effort) {
			if (effort == 0) {				// if command is STOP
				if (motorPins.PWM != 0xFF)
					analogWrite(motorPins.PWM, 0);	// set motor PWM state to 'stopped'
				if (motorPins.Up != 0xFF)
					digitalWrite(motorPins.Up, LOW);	// set the 'up' line to 'stopped'
				if (motorPins.Down != 0xFF)
					digitalWrite(motorPins.Down, LOW);// set the 'down' line to 'stopped'
					
				if (BrakingEnabled) {	// if braking enabled, engage brake
					if (motorPins.Polarity != 0xFF)
						digitalWrite(motorPins.Polarity, BrakePolarity);	// set polarity
					if (motorPins.Brake != 0xFF)
						digitalWrite(motorPins.Brake, HIGH);				// enable brake
					if ((motorPins.PWM != 0xFF) && motorPins.BrakePWM)
						analogWrite(motorPins.PWM, BrakePWM);				// enable port
				}
			} else {							// if command is MOVE/PULSE
				if (motorPins.PWM != 0xFF)
					analogWrite(motorPins.PWM, 0);		// set motor PWM state to 'stopped'
				if (motorPins.Brake != 0xFF)
					digitalWrite(motorPins.Brake, LOW);	// disengage brake
				if (motorPins.Polarity != 0xFF)
					digitalWrite(motorPins.Polarity, polarity ? HIGH : LOW); // set direction
				if (motorPins.PWM != 0xFF)
					analogWrite(motorPins.PWM, effort);	// set motor PWM state to 'go'
				if (motorPins.Up != 0xFF)
					digitalWrite(motorPins.Up, polarity ? HIGH : LOW);		// set 'up'
				if (motorPins.Down != 0xFF)
					digitalWrite(motorPins.Down, (!polarity) ? HIGH : LOW);	// set 'down'
			}
		}

	public:
		//
		//  MotorDC::MotorDC(...) - constructor; arguments are pin assignments
		//
		//  pwm       - Generates PWM output for the named port (A or B), with a duty
		//              cycle proportional to the commanded motor effort, e.g.,
		//              (0% effort = 0% duty cycle; 100% effort = 100% duty cycle).
		//              This line may generate a 100% duty cycle voltage when the
		//              motor is at rest if braking is enabled.  This pin MUST be a
		//              PWM-capable pin.
		//  brake     - Generates a HIGH signal for the named port, when braking
		//              enabled, and LOW otherwise.
		//  direction - Generates a HIGH signal for the named port to indicate
		//              a forward motor command, or a LOW signal to indicate a
		//              motor reverse.  Note that this line is always in one of
		//              these two states, even when a movement command is not active.
		//  up        - Set to HIGH when a FORWARD motor command is active, and LOW
		//              at all other times.
		//  down      - Set to HIGH when a REVERSE motor command is active, and LOW
		//              at all other times.
		//  sense     - Reads an A/D line to measure motor current.  The scale is
		//              currently scaled as per the Arduino Motor Shield specification.
		//  brakePWM  - Indicates that the PWM line should be asserted/high whenever
		//              the brake is enabled.  This should be true to support both
		//              IN1/IN2/ENABLE and PWM/DIR/BRAKE interfaces.
		//
		MotorDC(uint8_t pwm, uint8_t direction, uint8_t up, uint8_t down, uint8_t sense, uint8_t led, uint8_t brake, bool brakePWM) :
			motorPins(pwm, brake, sense, direction, up, down, led, brakePWM) {
			// start with a clean state
			PulseStart = PulseLength = 0UL;
			Current = 0.0;
			Pulsing = false;
			Running = false;
			BrakingEnabled = false;
		}

		// empty virtual dtor placeholder
		virtual ~MotorDC() { /* nop */ }

		//
		//  Setup() - initialize the hardware; this should be called once from setup()
		//
		virtual void Setup() {
			// configure the I/O pins
			if (motorPins.PWM != 0xFF) {
				pinMode(motorPins.PWM, OUTPUT);
				digitalWrite(motorPins.PWM, LOW);
			}
			if (motorPins.Brake != 0xFF) {
				pinMode(motorPins.Brake, OUTPUT);
				digitalWrite(motorPins.Brake, LOW);
			}
			if (motorPins.Polarity != 0xFF) {
				pinMode(motorPins.Polarity, OUTPUT);
				digitalWrite(motorPins.Polarity, LOW);
			}
			if (motorPins.Current != 0xFF) {
				pinMode(motorPins.Current, INPUT);
			}
			if (motorPins.Up != 0xFF) {
				pinMode(motorPins.Up, OUTPUT);
				digitalWrite(motorPins.Up, LOW);
			}
			if (motorPins.Down != 0xFF) {
				pinMode(motorPins.Down, OUTPUT);
				digitalWrite(motorPins.Down, LOW);
			}
			if (motorPins.LED != 0xFF) {
				pinMode(motorPins.LED, OUTPUT);
				digitalWrite(motorPins.LED, LOW);
			}

			// set motor state to 'stopped'
			Stop();

#ifdef DEBUG // print pin configuration for the port
			HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, Port = ")));
			HostSerial.print(i);
			if (motorPins.PWM != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", PWM = ")));
				HostSerial.print(motorPins.PWM);
			}
			if (motorPins.Brake != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Brake = ")));
				HostSerial.print(motorPins.Brake);
			}
			if (motorPins.Polarity != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Polarity = ")));
				HostSerial.print(motorPins.Polarity);
			}
			if (motorPins.Current != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Current = ")));
				HostSerial.print(motorPins.Current);
			}
			if (motorPins.Up != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Up = ")));
				HostSerial.print(motorPins.Up);
			}
			if (motorPins.Down != 0xFF) {
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", Down = ")));
				HostSerial.print(motorPins.Down);
			}
			HostSerial.println();
#endif
		}
		
		//
		//  Pulse(...) - pulse the motor with a specific direction, speed, and duration
		//
		virtual bool Pulse(bool direction, uint8_t speed, uint16_t howLong) {
			// command the motor
			CommandMotor(direction, speed);

			// record the pulse details
			Pulsing = true;				// store new pulse state state
			Running = false;			// cancel pending 'move'
			PulseStart = millis();		// get the start stamp
			PulseLength = howLong;		// compute pulse end time

			// success
			return true;
		}
		
		//
		//  Step(...) - move step-equipped motor specific number of steps at given rate
		//
		virtual bool Step(bool direction, uint8_t speed, uint16_t steps) {
			// stepping isn't supported
			return false;
		}

		//
		//  Move(...) - move the motor continuously with a specific direction and speed
		//
		virtual bool Move(bool direction, uint8_t speed) {
			Pulsing = false;
			Running = (speed != 0);
			CommandMotor(direction, speed);

			// success
			return true;
		}

		//
		//  Run() - update state, timers, actions, etc.; this should be called once per loop() iteration
		//
		virtual void Run() {
			unsigned long now = millis();
			
			// if pulse in progress
			if (Pulsing && (Elapsed(now, PulseStart) >= PulseLength)) {
				Pulsing = false;	// cancel pending pulse
				Running = false;	// cancel pending move

				// command the motor to STOP
				Stop();
			}
		}
		
		//
		//  Stop() - stop the motor, regardless of what it is doing
		//
		virtual bool Stop() {
			// CommandMotor('stop') will also apply the brake
			CommandMotor(false, 0);

			// clear any running flags
			Pulsing = Running = false;

			// success
			return true;
		}
		
		//
		//  ReadCurrent() - read the motor current and return it
		//
		virtual float ReadCurrent() {
			if (motorPins.Current != 0xFF)
				SetCurrent(analogRead(motorPins.Current));
			else
				Current = 0;
			return Current;
		}
		
		//
		//  GetCurrent() - return the last current reading
		//
		float GetCurrent() const {
			return Current;
		}
		
		//
		//  LED(state) - update the 'port selected' LED state
		//
		virtual void LED(bool state) {
			digitalWrite(motorPins.LED, state);
		}
		
		//
		//  SetBrake(state) - set state of braking feature
		//
		virtual bool SetBrake(bool state) {
			// save the flag; applies only to this port
			BrakingEnabled = state;

			// if the motor is idle, apply the new braking state now
			if (!Running && !Pulsing) {
				CommandMotor(false, 0); // which stops AND brakes the motor
#ifdef DEBUG
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, Port was idle, so new value of ")));
				HostSerial.print(BrakingEnabled ? strcpy_P(ioBuffer, PSTR("ENABLED")) : strcpy_P(ioBuffer, PSTR("DISABLED")));
				HostSerial.println(strcpy_P(ioBuffer, PSTR(" was applied immediately.")));
#endif
			}
#ifdef DEBUG
			else { // just mention it to the debug output, setting will take effect when motor stops
				HostSerial.print(strcpy_P(ioBuffer, PSTR("#debug, Braking for port ")));
				HostSerial.print(port);
				HostSerial.print(strcpy_P(ioBuffer, PSTR(", now = ")));
				HostSerial.println(BrakingEnabled ? strcpy_P(ioBuffer, PSTR("ENABLED")) : strcpy_P(ioBuffer, PSTR("DISABLED")));
			}
#endif
			// success
			return true;
		}


		//
		//  GetBrake() - get the brake setting
		//
		virtual bool GetBrake() {
			return BrakingEnabled;
		}

		//
		//  Read the current stepper position; return true if supported
		//
		virtual bool GetPosition(long *outValue) {
			// not supported
			return false;
		}
		
		//
		//  Reset the current stepper position to zero; return true if supported
		//
		virtual bool ZeroPosition(void) {
			// not supported
			return false;
		}

		//
		//  Seek to a specific position
		//
		virtual bool Seek(long target, uint8_t speed) {
			// not supported
			return false;
		}

		//
		//  Get an enable pin definition
		//
		virtual EnablePin *GetEnable(uint8_t idx) {
			// not supported
			return 0;
		}

		//
		//  Returns true iff the motor is moving
		//
		virtual bool IsRunning() const {
			return Running || Pulsing;
		}
};

#endif // AUTOCAP_MOTORDC_H

