/*
 *
 *
 *    State.h
 *
 *    AutoCap Arduino Firmware
 *    Version 1.5
 *
 *    State machine storage block.
 *
 *    This software is intended for use with the AutoCap software core,
 *    to auto-tune a loop.  This version of the firmware provides basic
 *    motor control, under control of the core, via the Arduino's host
 *    serial link.
 *
 *    License: GNU General Public License Version 3.0.
 *    
 *    Copyright (C) 2014-2016 by Matthew K. Roberts, KK5JY.
 *    All rights reserved.
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *    
 *    This program is distributed in the hope that it will be useful, but
 *    WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    General Public License for more details.
 *    
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see: http://www.gnu.org/licenses/
 *    
 *
 */

#ifndef __AUTOCAP_STATE_H
#define __AUTOCAP_STATE_H

enum InputModes {
	Invalid = 0,

	// movement modes
	Momentary = 1,  // press-and-hold
	Continuous = 2, // latched
	DoPulse = 3,
	#if USE_ACCELSTEPPER == true
	DoStep = 9,
	#endif

	// configuration modes
	SetPulse = 10,
	#if USE_ACCELSTEPPER == true
	SetSteps = 11,
	#endif
	SetPWM = 12,
};

//
//  State block
//
struct State {
	public: // constants
		const uint8_t PortCount; // set by ctor

		const uint16_t MinWidth = 10; // ms
		const uint16_t MaxWidth = 1000; // ms

		const uint16_t MinSteps = 1; // steps
		const uint16_t MaxSteps = 128; // steps

		const uint16_t WidthCoarse = 50; // ms
		const uint16_t WidthFine = 10; // ms

		const uint16_t PWMCoarse = 5; // %

	public: // public properties
		InputModes InputMode;

		uint16_t Width;   // the pulse width (ms)
		uint16_t Steps;   // the step size (steps)
		uint8_t UserPWM;  // the PWM value (1-100)
		uint8_t UserPort; // the user selected port

		uint64_t LastPulse; // the last pulse command

	private: // internal stuff
		IMotor * const * Motors;
		unsigned long modePressed;
		bool LastMotorState;

	private:
		void Assign(const State &s) {
			InputMode = s.InputMode;
			Width = s.Width;
#if USE_ACCELSTEPPER == true
			Steps = s.Steps;
#endif
			UserPWM = s.UserPWM;
			UserPort = s.UserPort;
		}

	public:
		State(IMotor * const motors[], uint8_t port_ct)
			: Motors(motors), PortCount(port_ct), LastMotorState(false) {
			InputMode = Momentary;
			Width = 100;
			Steps = 4;
			UserPWM = 100;
			UserPort = 0;

			modePressed = 0;
		}

		State(const State &s)
			: Motors(s.Motors), PortCount(s.PortCount), LastMotorState(false) {
			Assign(s);

			modePressed = 0;
		}

	public:
		State &operator = (const State &s) {
			Assign(s);
			return *this;
		}

		bool Equals(const State &s) const {
			return
				(InputMode == s.InputMode) &&
				(Width == s.Width) &&
				(Steps == s.Steps) &&
				(UserPWM == s.UserPWM) &&
				(UserPort == s.UserPort);
		}

		/*
		bool SettingsDifferent(const State &s) const {
			return
				(InputMode != s.InputMode) ||
				(Width != s.Width) ||
				(Steps != s.Steps) ||
				(UserPWM != s.UserPWM) ||
				(UserPort != s.UserPort);
		}
		*/

		bool operator == (const State &s) const {
			return Equals(s);
		}

		bool operator != (const State &s) const {
			return !Equals(s);
		}

	public:
		/*
		bool SettingActive() const {
			return InputMode >= 10;
		}
		*/

		void RecordPulse() {
			LastPulse = millis();
			if (LastPulse == 0) {
				LastPulse = 1;
			}
		}

		// user commands
		void Up(bool pressed);
		void Down(bool pressed);
		void Stop(bool pressed);
		void Mode();
		void ModeHold();
		void PortSelect();

		// return true iff attached displays should be updated
		bool UpdateDisplays(unsigned long now);
};

bool State::UpdateDisplays(unsigned long now = millis()) {
	bool result = false;
	if (LastPulse != 0) {
		result = (Elapsed(now, LastPulse) >= Width);
		if (result) {
			LastPulse = 0;
		}
	}

	// read motor state
	bool isRunning = motors[UserPort]->IsRunning();
	result |= (isRunning != LastMotorState);
	LastMotorState = isRunning;

	return result;
}

inline void State::PortSelect() {
	UserPort = (UserPort + 1) % PortCount;
	// update motor LEDs
	for (int i = 0; i != PortCount; ++i) {
		motors[i]->LED(i == UserPort);
	}
}

static inline uint8_t ToRegister(uint8_t normalized) {
	return (normalized * 255) / 100;
}

inline void State::Up(bool pressed) {
	const uint8_t polarity = 1;

	if (pressed) {
		//
		//  button-down
		//
		switch (InputMode) {
			case Momentary:
				Motors[UserPort]->Move(polarity, ToRegister(UserPWM));
				break;
			case Continuous:
				Motors[UserPort]->Move(polarity, ToRegister(UserPWM));
				break;
			case DoPulse:
				Motors[UserPort]->Pulse(polarity, ToRegister(UserPWM), Width);
				RecordPulse();
				break;
			case SetPulse:
				if (Width == MaxWidth) {
					return;
				}
				if (Width >= 100) {
					Width += WidthCoarse;
				} else {
					Width += WidthFine;
				}
				break;
			case SetPWM:
				if (UserPWM == 100)
					return;
				UserPWM += PWMCoarse;
				break;
#if USE_ACCELSTEPPER == true
			case DoStep:
				Motors[UserPort]->Step(polarity, ToRegister(UserPWM), Steps);
				break;
			case SetSteps:
				if (Steps == MaxSteps) {
					return;
				}
				Steps *= 2;
				break;
#endif
			default:
				break;
		}
	} else {
		//
		//  button-up
		//
		switch (InputMode) {
			case Momentary:
				Motors[UserPort]->Stop();
				break;
			default:
				// nop
				break;
		}
	}
}

inline void State::Down(bool pressed) {
	const uint8_t polarity = 0;

	if (pressed) {
		//
		//  button-down
		//
		switch (InputMode) {
			case Continuous:
				Motors[UserPort]->Move(polarity, ToRegister(UserPWM));
				break;
			case Momentary:
				Motors[UserPort]->Move(polarity, ToRegister(UserPWM));
				break;
			case DoPulse:
				Motors[UserPort]->Pulse(polarity, ToRegister(UserPWM), Width);
				RecordPulse();
				break;
			case SetPulse:
				if (Width == MinWidth) {
					return;
				}
				if (Width > 100) {
					Width -= WidthCoarse;
				} else {
					Width -= WidthFine;
				}
				break;
			case SetPWM:
				if (UserPWM == PWMCoarse)
					return;
				UserPWM -= PWMCoarse;
				break;
#if USE_ACCELSTEPPER == true
			case DoStep:
				Motors[UserPort]->Step(polarity, ToRegister(UserPWM), Steps);
				break;
			case SetSteps:
				if (Steps == MinSteps) {
					return;
				}
				Steps /= 2;
				break;
#endif
			default:
				break;
		}
	} else {
		//
		//  button-up
		//
		switch (InputMode) {
			case Momentary:
				Motors[UserPort]->Stop();
				break;
			default:
				// nop
				break;
		}
	}
}

inline void State::Stop(bool pressed) {
	Motors[UserPort]->Stop();
}

inline void State::ModeHold() {
	//
	//  These are the transitions of operation <--> SET
	//
	switch (InputMode) {
		case Momentary: // momentary --> setPWM
			InputMode = SetPWM;
			break;
		case Continuous: // continuous --> setPWM
			InputMode = SetPWM;
			break;
		case DoPulse:  // pulse --> set-pulse
			InputMode = SetPulse;
			break;
		case SetPulse: // set-pulse --> pulse
			InputMode = DoPulse;
			break;
#if USE_ACCELSTEPPER == true
		case DoStep: // step -> set-step
			InputMode = SetSteps;
			break;
		case SetSteps: // set-step -> step
			InputMode = DoStep;
			break;
#endif
		default: // set-pwm -> momentary
			InputMode = Momentary;
			break;
	}
}

inline void State::Mode() {
	switch (InputMode) {
		case SetPWM:
			InputMode = SetPulse;
			break;
		case SetPulse:
#if USE_ACCELSTEPPER == true
			InputMode = SetSteps;
			break;
		case SetSteps:
#endif
			InputMode = SetPWM;
			break;
		case Momentary:
			InputMode = Continuous;
			break;
		case Continuous:
			InputMode = DoPulse;
			break;
		case DoPulse:
#if USE_ACCELSTEPPER == true
			InputMode = DoStep;
			break;
		case DoStep:
			InputMode = Momentary;
			break;
#endif
		default:
			InputMode = Momentary;
			break;
	}
}
#endif
